/*
  The ls_sless plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The ls_sless plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The ls_sless plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include <partysip/psp_utils.h>

#include <ppl/ppl_dns.h>

#include "ls_sless.h"

ls_sless_ctx_t *ls_sless_context = NULL;

extern psp_plugin_t PPL_DECLARE_DATA ls_sless_plugin;

/*
  This plugin can be configured to:
  * act as a redirect server.

  * mandate stateless for INVITE with an IP in the host of the request-uri
  * mandate stateless for INVITE with an IP in the host of the request-uri
  * mandate stateless for INVITE with a FQDN in the host of the request-uri

  * mandate stateless for REQUEST with an IP in the host of the request-uri
  * mandate stateless for REQUEST with an IP in the host of the request-uri
  * mandate stateless for REQUEST with a FQDN in the host of the request-uri

  * mandate record-routing to stay on the path for the duration of the call.
       ("proxy-mode")

  NOTE: YOU SHOULD REALLY USE THE CONFIGURATION SAMPLE PROVIDED BELOW.
  For design reason, this is not a good idea to mandate stateless behavior
  when an FQDN is met. But it's a good practive to do so if you find a
  IP... (so the processing will only take a few time).


  Why allowing FQDN resolution with stateless behavior??
  JUST FOR TEST PURPOSE!! I have warned you about it!
  */

/*
  Configuration sample for INVITE processing.
  
  <sless>


  # invite rquri ip reject
  invite rquri ip forward
  
  invite rquri fqdn reject
  # invite rquri fqdn forward
  
  # invite header route reject
  invite rquri route forward
  </sless>
*/
#define INV_RQURI_IP       0x001
#define INV_RQURI_FQDN     0x002
#define INV_HEADER_ROUTE   0x004

#define ISSET_INV_RQURI_IP(flag)     ((~flag|~INV_RQURI_IP)==~INV_RQURI_IP)
#define ISSET_INV_RQURI_FQDN(f)      ((~f|~INV_RQURI_FQDN)==~INV_RQURI_FQDN)
#define ISSET_INV_HEADER_ROUTE(flag) ((~flag|~INV_HEADER_ROUTE)==~INV_HEADER_ROUTE)

/*
#define SET_INV_RQURI_IP(flag)     (flag=(flag|INV_RQURI_IP)
#define SET_INV_RQURI_FQDN(flag)   (flag=((flag&~INV_RQURI_FQDN)|INV_RQURI_FQDN))
#define SET_INV_HEADER_ROUTE(flag) (flag=((flag&~INV_HEADER_ROUTE)|INV_HEADER_ROUTE))
*/
#define SET_INV_RQURI_IP(flag)     (flag=flag|INV_RQURI_IP)
#define SET_INV_RQURI_FQDN(flag)   (flag=flag|INV_RQURI_FQDN)
#define SET_INV_HEADER_ROUTE(flag) (flag=flag|INV_HEADER_ROUTE)

/*
  Configuration sample for REQUEST processing.

  <sless>

  # request rquri ip ignore
  request rquri ip forward
  
  request rquri fqdn ignore
  # request rquri fqdn forward
  
  # request header route ignore
  request rquri route forward
  </sless>
*/
#define REQ_RQURI_IP      0x10
#define REQ_RQURI_FQDN    0x20
#define REQ_HEADER_ROUTE  0x40


#define ISSET_REQ_RQURI_IP(flag)     ((~flag|~REQ_RQURI_IP)==~REQ_RQURI_IP)
#define ISSET_REQ_RQURI_FQDN(f)      ((~f|~REQ_RQURI_FQDN)==~REQ_RQURI_FQDN)
#define ISSET_REQ_HEADER_ROUTE(flag) ((~flag|~REQ_HEADER_ROUTE)==~REQ_HEADER_ROUTE)

#define SET_REQ_RQURI_IP(flag)       (flag=flag|REQ_RQURI_IP)
#define SET_REQ_RQURI_FQDN(flag)     (flag=flag|REQ_RQURI_FQDN)
#define SET_REQ_HEADER_ROUTE(flag)   (flag=flag|REQ_HEADER_ROUTE)

/*
  Configuration sample for REQUEST processing.

  <sless>
  # mode redirect
  mode stateless

  # enable record routing (for invite only?)
  record-route on
  </sless>
*/
#define REDIRECT_MODE      0x80
#define R_ROUTE_MODE       0x08

#define ISSET_REDIRECT_MODE(flag)    ((~flag|~REDIRECT_MODE)==~REDIRECT_MODE)
#define SET_REDIRECT_MODE(flag)      (flag=flag|REDIRECT_MODE)

#define ISSET_R_ROUTE_MODE(flag)     ((~flag|~R_ROUTE_MODE)==~R_ROUTE_MODE)
#define SET_R_ROUTE_MODE(flag)       (flag=flag|R_ROUTE_MODE)

int
ls_sless_ctx_init ()
{
  config_element_t *elem;
  int i;

  ls_sless_context = (ls_sless_ctx_t *) smalloc (sizeof (ls_sless_ctx_t));
  if (ls_sless_context == NULL)
    return -1;

  ls_sless_context->flag = 0;
  elem = psp_config_get_sub_element ("mode", "sless", NULL);
  if (elem == NULL || elem->value == NULL)
    {
  } else if (0 == strcmp (elem->value, "redirect"))
    SET_REDIRECT_MODE (ls_sless_context->flag);
  else if (0 == strcmp (elem->value, "stateless"))
    {
  } else
    goto lsci_error1;           /* error, bad option */

  elem = psp_config_get_sub_element ("record-route", "sless", NULL);
  if (elem == NULL || elem->value == NULL)
    {
  } else if (0 == strcmp (elem->value, "off"))
    {
  } else if (0 == strcmp (elem->value, "on"))
    SET_R_ROUTE_MODE (ls_sless_context->flag);
  else
    goto lsci_error1;           /* error, bad option */

  /* load the configuration behavior for INVITE */
  i = ls_sless_load_invite_config ();
  if (i != 0)
    goto lsci_error2;

  i = ls_sless_load_request_config ();
  if (i != 0)
    goto lsci_error2;

  if (ISSET_R_ROUTE_MODE (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to do record-routing!\n"));
    }

  if (ISSET_REDIRECT_MODE (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured in redirect mode!\n"));
    }
  if (ISSET_INV_RQURI_IP (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward INV with an IP in the rquri!\n"));
    }
  if (ISSET_INV_RQURI_FQDN (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward INV with a FQDN in the rquri!\n"));
    }
  if (ISSET_INV_HEADER_ROUTE (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward INV with a Route header!\n"));
    }
  if (ISSET_REQ_RQURI_IP (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward REQ with an IP in the rquri!\n"));
    }
  if (ISSET_REQ_RQURI_FQDN (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward REQ with a FQDN in the rquri!\n"));
    }
  if (ISSET_REQ_HEADER_ROUTE (ls_sless_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_sless plugin: configured to forward REQ with a Route header!\n"));
    }

  return 0;

lsci_error2:
lsci_error1:
  sfree (ls_sless_context);
  ls_sless_context = NULL;
  return -1;
}

int
ls_sless_load_invite_config ()
{
  config_element_t *elem;
  config_element_t *next_elem;
  char *buf;
  char *next;
  char *type;
  char *tmp;
  int i;

  elem = psp_config_get_sub_element ("invite", "sless", NULL);

  while (elem != NULL)
    {
      /*
         invite [rquri | header] [ip | fqdn] [forward|ignore]
       */
      i = psp_util_get_and_set_next_token (&type, elem->value, &buf);
      if (i != 0 || type == NULL)
        return -1;
      sclrspace (type);
      i = psp_util_get_and_set_next_token (&tmp, buf, &next);
      if (i != 0 || tmp == NULL)
        {
          sfree (type);
          return -1;
        }
      sclrspace (tmp);
      sclrspace (next);

      if (0 == strcmp (type, "rquri"))
        {
          /* next token is 'ip' or 'fqdn' */
          if (0 == strcmp (next, "forward"))
            {
              if (0 == strcmp (tmp, "ip"))
                {
                  SET_INV_RQURI_IP (ls_sless_context->flag);
              } else if (0 == strcmp (tmp, "fqdn"))
                {
                  SET_INV_RQURI_FQDN (ls_sless_context->flag);
                }
            }
      } else if (0 == strcmp (type, "header"))
        {
          /* next token is 'ip' or 'fqdn' */
          if (0 == strcmp (next, "forward"))
            {
              if (0 == strcmp (tmp, "route"))
                {
                  SET_INV_HEADER_ROUTE (ls_sless_context->flag);
                }
            }
      } else
        return -1;
      sfree (type);
      sfree (tmp);
      next_elem = elem;
      if (next_elem == NULL)
        return 0;
      elem = psp_config_get_sub_element ("invite", "sless", next_elem);
    }
  return 0;
}

int
ls_sless_load_request_config ()
{
  config_element_t *next_elem;
  config_element_t *elem;
  char *buf;
  char *next;
  char *type;
  char *tmp;
  int i;

  elem = psp_config_get_sub_element ("request", "sless", NULL);

  while (elem != NULL)
    {
      /*
         request [rquri | header] [ip | fqdn] [forward|ignore]
       */
      i = psp_util_get_and_set_next_token (&type, elem->value, &buf);
      if (i != 0 || type == NULL)
        return -1;
      sclrspace (type);
      i = psp_util_get_and_set_next_token (&tmp, buf, &next);
      if (i != 0 || tmp == NULL)
        {
          sfree (type);
          return -1;
        }
      sclrspace (tmp);
      sclrspace (next);

      if (0 == strcmp (type, "rquri"))
        {
          /* next token is 'ip' or 'fqdn' */
          if (0 == strcmp (next, "forward"))
            {
              if (0 == strcmp (tmp, "ip"))
                {
                  SET_REQ_RQURI_IP (ls_sless_context->flag);
              } else if (0 == strcmp (tmp, "fqdn"))
                {
                  SET_REQ_RQURI_FQDN (ls_sless_context->flag);
                }
            }
      } else if (0 == strcmp (type, "header"))
        {
          /* next token is 'ip' or 'fqdn' */
          if (0 == strcmp (next, "forward"))
            {
              if (0 == strcmp (tmp, "route"))
                {
                  SET_REQ_HEADER_ROUTE (ls_sless_context->flag);
                }
            }
        }
      sfree (type);
      sfree (tmp);
      next_elem = elem;
      if (next_elem == NULL)
        return 0;
      elem = psp_config_get_sub_element ("request", "sless", next_elem);
    }
  return 0;
}

void
ls_sless_ctx_free ()
{
  if (ls_sless_context == NULL)
    return;

  sfree (ls_sless_context);
  ls_sless_context = NULL;
}

/* HOOK METHODS */

/*
  This method returns:
  -2 if plugin consider this request should be totally discarded!
  -1 on error
  0  nothing has been done
  1  things has been done on psp_req element
*/
int
cb_ls_sless_search_location (psp_req_t * psp_req)
{
  location_t *loc;
  url_t *url;
  int is_an_ip;
  route_t *route;
  int i;

  /* default OUTPUT */
  if (ISSET_R_ROUTE_MODE (ls_sless_context->flag))
    psp_req_set_property (psp_req, PSP_STAY_ON_PATH);
  else
    psp_req_set_property (psp_req, 0);

  if (!ISSET_REDIRECT_MODE (ls_sless_context->flag))
    /* state less or default mode */
    psp_req_set_mode (psp_req, PSP_SLESS_MODE);
  else
    {
      psp_req_set_uas_status (psp_req, 302);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
    }
  psp_req_set_state (psp_req, PSP_MANDATE);

  /* if the request contains a route header, we don't need to perform
     search location any more... */
  msg_getroute (psp_req->request, 1, &route);
  if (route != NULL)
    {
      psp_req_set_mode (psp_req, PSP_SLESS_MODE);       /* MUST BE STATELESS HERE */
      if (MSG_IS_INVITE (psp_req->request))
        {
          if (!ISSET_INV_HEADER_ROUTE (ls_sless_context->flag))
            {
              psp_req_set_state (psp_req, PSP_PROPOSE); /* do not mandate */
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                      "ls_sless plugin: ignore this request\n"));
          } else
            {
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                                      "ls_sless plugin: mandate stateless (or redirect) mode for request\n"));
            }
      } else
        {
          if (!ISSET_REQ_HEADER_ROUTE (ls_sless_context->flag))
            {
              psp_req_set_state (psp_req, PSP_PROPOSE); /* do not mandate */
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                      "ls_sless plugin: ignore this request.\n"));
          } else
            {
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                                      "ls_sless plugin: mandate stateless (or redirect) mode for request.\n"));
            }
        }
      return 0;
    }

  /* in the case where we detect an IP address in the request-URI
     which is not the local proxy, then we should forward it to
     this destination. */
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                          "ls_sless plugin: checking if we are responsible for request-URI '%s'\n",
                          psp_req->request->strtline->rquri->host));
  if (0 ==
      psp_core_is_responsible_for_this_request_uri (psp_req->request->strtline->
                                                    rquri))
    {
      psp_req_set_state (psp_req, PSP_CONTINUE);
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                              "ls_sless plugin: ignore this request which belongs to this domain.\n"));
      return 0;
    }

  /* handle cases where a FQDN is found */
  if (MSG_IS_INVITE (psp_req->request))
    {
      is_an_ip = inet_addr (psp_req->request->strtline->rquri->host);
      if (is_an_ip == -1 && !ISSET_INV_RQURI_FQDN (ls_sless_context->flag))
        {
          /* ignore this request as it needs a request resolution */
          psp_req_set_state (psp_req, PSP_CONTINUE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                  "ls_sless plugin: ignore this request which needs an address resolution.\n"));
          return 0;
      } else if (is_an_ip != -1 && !ISSET_INV_RQURI_IP (ls_sless_context->flag))
        {
          /* ignore this request. THIS SHOULD BE NEVER CONFIGURED THIS WAY
             except if you know what you are doing... */
          psp_req_set_state (psp_req, PSP_CONTINUE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                  "ls_sless plugin: ignore this request which needs an address resolution.\n"));
          return 0;
        }
  } else                        /* other request */
    {
      is_an_ip = inet_addr (psp_req->request->strtline->rquri->host);
      if (is_an_ip == -1 && !ISSET_REQ_RQURI_FQDN (ls_sless_context->flag))
        {
          /* ignore this request as it needs a request resolution */
          psp_req_set_state (psp_req, PSP_CONTINUE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                  "ls_sless plugin: ignore this request which need an address resolution.\n"));
          return 0;
        }
      if (is_an_ip != -1 && !ISSET_REQ_RQURI_IP (ls_sless_context->flag))
        {
          /* ignore this request. THIS SHOULD BE NEVER CONFIGURED THIS WAY
             except if you know what you are doing... */
          psp_req_set_state (psp_req, PSP_CONTINUE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                  "ls_sless plugin: ignore this request which need an address resolution.\n"));
          return 0;
        }
    }

  if (is_an_ip == -1)
    /* even if we need to resolve an FQDN address,
       we still mandate for stateless... MAINLY THIS IS FOR TESTING PURPOSE */
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                              "WARNING: ls_sless plugin: request accepted but it needs an address resolution! Use this configuration mode at your own risk.\n"));
    }

  /* we MUST add ONE a location in psp_req */
  i = url_clone (psp_req->request->strtline->rquri, &url);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "ls_sless plugin: Could not clone request-uri!\n"));
      psp_req_set_uas_status (psp_req, 400);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_state (psp_req, PSP_MANDATE);
      return -1;                /* error case (process can continue??) */
    }
  i = location_init (&loc, url, 3600);
  if (i != 0)
    {                           /* This can only happen in case we don't have enough memory */
      url_free (url);
      sfree (url);
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_BUG, NULL,
                              "ls_sless plugin: Could not create location info!\n"));
      psp_req_set_uas_status (psp_req, 400);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_state (psp_req, PSP_MANDATE);
      return -1;                /* error case (process can continue...) */
    }
  ADD_ELEMENT (psp_req->locations, loc);
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                          "ls_sless plugin: mandate stateless (or redirect) mode for request.\n"));
  return 0;
}
