/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>

#include <partysip/psp_config.h>
#include <psp_utils.h>

#include <ppl/ppl_init.h>
#include <ppl/ppl_getopt.h>

#define PSP_SERVER_BASEARGS "f:l:d:vVilph?X"

#define PSP_SERVER_BASEVENDOR "WellX Telecom"
#define PSP_SERVER_BASEPRODUCT "Partysip"

#ifdef WIN32
#define PSP_SERVER_BASEREVISION PSP_VERSION
#define PSP_SERVER_BASEREVISION PSP_VERSION
#else
#define PSP_SERVER_BASEREVISION VERSION
#define PSP_SERVER_BASEREVISION VERSION
#endif

#define PSP_SERVER_BASEVERSION PSP_SERVER_BASEPRODUCT "/" PSP_SERVER_BASEREVISION
#define PSP_SERVER_VERSION  PSP_SERVER_BASEVERSION

#define PSP_SERVER_INSTALL_DIR PSP_SERVER_PREFIX

#if defined(__DATE__) && defined(__TIME__)
static const char server_built[] = __DATE__ " " __TIME__;
#else
static const char server_built[] = "unknown";
#endif

#define PSP_SERVER_WHO_AM_I PSP_SERVER_BASEVENDOR "/" PSP_SERVER_VERSION

#define PSP_PLUGIN_API "O.1"

/* verify the syntax of the server header in SIP ... */
#define PSP_SERVER_HEADER   "WellX/partysip-" PSP_SERVER_BASEREVISION

FILE *log_file = NULL;


/* Developper can grab the compile flag here... */
static void
show_compile_settings ()
{
  printf ("Server header:      %s\n", PSP_SERVER_HEADER);
  printf ("Server version:     %s\n", PSP_SERVER_BASEVERSION);
  printf ("Global Config File  %s\n", PSP_SERVER_INSTALL_DIR);
  printf ("Server built:       %s\n", server_built);
  printf ("Server's interface: %s\n", PSP_PLUGIN_API);
  printf ("Architecture:       %ld-bit\n", 8 * (long) sizeof (void *));
  printf ("Server compiled with....\n");
#ifdef STDC_HEADERS
  printf (" -DSTDC_HEADERS\n");
#endif
#ifdef HAVE_ASSERT_H
  printf (" -DHAVE_ASSERT_H\n");
#endif
#ifdef HAVE_CTYPE_H
  printf (" -DHAVE_CTYPE_H\n");
#endif
#ifdef HAVE_DL_H
  printf (" -DHAVE_DL_H\n");
#endif
#if HAVE_DLFCN_H
  printf (" -DHAVE_DLFCN_H\n");
#endif
#if HAVE_FCNTL_H
  printf (" -DHAVE_FCNTL_H\n");
#endif
#if HAVE_MALLOC_H
  printf (" -DHAVE_MALLOC_H\n");
#endif
#ifdef HAVE_PTH_PTHREAD_H
  printf (" -DHAVE_PTH_PTHREAD_H\n");
#endif
#ifdef HAVE_PTHREAD_H
  printf (" -DHAVE_PTHREAD_H\n");
#endif
#if HAVE_SEMAPHORE_H
  printf (" -DHAVE_SEMAPHORE_H\n");
#endif
#if HAVE_SIGNAL_H
  printf (" -DHAVE_SIGNAL_H\n");
#endif
#if HAVE_STDARG_H
  printf (" -DHAVE_STDARG_H\n");
#endif
#if HAVE_STDDEF_H
  printf (" -DHAVE_STDDEF_H\n");
#endif
#if HAVE_STDIO_H
  printf (" -DHAVE_STDIO_H\n");
#endif
#if HAVE_STDLIB_H
  printf (" -DHAVE_STDLIB_H\n");
#endif
#if HAVE_STRING_H
  printf (" -DHAVE_STRING_H\n");
#endif
#ifdef HAVE_STRINGS_H
  printf (" -DHAVE_STRINGS_H\n");
#endif
#if HAVE_SYS_SEM_H
  printf (" -DHAVE_SYS_SEM_H\n");
#endif
#ifdef HAVE_SYS_SIGNAL_H
  printf (" -DHAVE_SYS_SIGNAL_H\n");
#endif
#ifdef HAVE_SYS_SOCKET_H
  printf (" -DHAVE_SYS_SOCKET_H\n");
#endif
#ifdef HAVE_SYS_TIME_H
  printf (" -DHAVE_SYS_TIME_H\n");
#endif
#if HAVE_SYS_TYPES_H
  printf (" -DHAVE_SYS_TYPES_H\n");
#endif
#ifdef HAVE_TIME_H
  printf (" -DHAVE_TIME_H\n");
#endif
#ifdef HAVE_UNISTD_H
  printf (" -DHAVE_UNISTD_H\n");
#endif
#ifdef HAVE_VARARGS_H
  printf (" -DHAVE_VARARGS_H\n");
#endif
  printf (" -DPSP_SERVER_INSTALL_DIR=%s\n", PSP_SERVER_INSTALL_DIR);
}

void
usage (int code)
{
  printf ("\n\
usage:\n\
\n\
   partysip [-vV] [-f config] [-d level -l logfile]\n\
\n\
   [-h]                  print this help\n\
   [-i]                  interactive mode\n\
   [-f config]           configuration file for partysip.\n\
   [-d level]            be verbose. 0 -> no output. 6 -> all output .\n\
   [-l logfile]          specify the log file.\n\
   [-v]                  print partysip info\n\
   [-V]                  print full partysip info with compile flags\n\n");
  exit (code);
}

static void
partysip_exit (int process_exit_value)
{

  psp_core_free ();
  psp_config_unload ();

  ppl_init_close ();

  if (log_file != NULL && log_file != stdout)
    fclose (log_file);

  exit (process_exit_value);
}

void
partysip_level (char *tmp)
{
  if (0 == strncmp (tmp, "0", 1))
    {
      TRACE_ENABLE_LEVEL (0);
    }
  if (0 == strncmp (tmp, "1", 1))
    {
      TRACE_ENABLE_LEVEL (1);
    }
  if (0 == strncmp (tmp, "2", 1))
    {
      TRACE_ENABLE_LEVEL (2);
    }
  if (0 == strncmp (tmp, "3", 1))
    {
      TRACE_ENABLE_LEVEL (3);
    }
  if (0 == strncmp (tmp, "4", 1))
    {
      TRACE_ENABLE_LEVEL (4);
    }
  if (0 == strncmp (tmp, "5", 1))
    {
      TRACE_ENABLE_LEVEL (5);
    }
  if (0 == strncmp (tmp, "d", 1))
    {
      TRACE_DISABLE_LEVEL (0);
      TRACE_DISABLE_LEVEL (1);
      TRACE_DISABLE_LEVEL (2);
      TRACE_DISABLE_LEVEL (3);
      TRACE_DISABLE_LEVEL (4);
      TRACE_DISABLE_LEVEL (5);
    }
}

#ifdef WIN32
char *
simple_readline (int descr)
{
  int i = 0;
  char *tmp = (char *) smalloc (201);

  fgets (tmp, 200, stdin);
  sclrspace (tmp);
  return tmp;
}
#else
char *
simple_readline (int descr)
{
  int ret;
  fd_set fset;

  FD_ZERO (&fset);
  FD_SET (descr, &fset);
  ret = select (descr + 1, &fset, NULL, NULL, NULL);

  if (FD_ISSET (descr, &fset))
    {
      char *tmp;
      int i;

      tmp = (char *) smalloc (201);
      i = read (descr, tmp, 200);
      tmp[i] = '\0';
      if (i > 0)
        sclrspace (tmp);
      return tmp;
    }
  return NULL;
}
#endif


void
main_run ()
{
  char *tmp;

  while (1)
    {
      printf ("(partysip)");
      fflush (stdout);
      tmp = simple_readline (1);
      if (tmp != NULL)
        {
          if (strlen (tmp) >= 6 && 0 == strncmp (tmp, "plugin", 6))
            ;
          else if (strlen (tmp) >= 4 && 0 == strncmp (tmp, "help", 4))
            ;
          else if (strlen (tmp) == 1 && 0 == strncmp (tmp, "q", 1))
            {
              sfree (tmp);
              partysip_exit (0);
          } else if (strlen (tmp) == 4 && 0 == strncmp (tmp, "quit", 4))
            {
              sfree (tmp);
              partysip_exit (0);
          } else
            printf ("error: %s: command not found!\n", tmp);
          sfree (tmp);
      } else
        partysip_exit (1);
    }

}

int
main_load_plugins ()
{
  int i;
  psp_plugin_t *psp_plugin;

  i = psp_plugin_load (&psp_plugin, "udp");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load UDP plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "syntax");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load SYNTAX plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "filter");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load FILTER plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "auth");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load AUTH plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "rgstrar");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load RGSTRAR plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "ls_sless");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load LS_SLESS plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "ls_localdb");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load LS_LOCALDB plugins!\n");
      return -1;
    }
  i = psp_plugin_load (&psp_plugin, "ls_sfull");
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load LS_SFULL plugins!\n");
      return -1;
    }
  return 0;
}

int
main (int argc, const char *const argv[])
{
  char c;
  int i;
  ppl_getopt_t *opt;
  ppl_status_t rv;
  const char *cf_config_file =
    PSP_SERVER_INSTALL_DIR "/etc/partysip/partysip.conf";
  int cf_debug_level = 0;
  const char *cf_log_file = NULL;
  int interactive_mode = 0;
  const char *optarg;

  /* ppl_initialize(); */
  ppl_init_open ();

  if (argc > 1 && strlen (argv[1]) == 1 && 0 == strncmp (argv[1], "-", 2))
    usage (0);
  if (argc > 1 && strlen (argv[1]) >= 2 && 0 == strncmp (argv[1], "--", 2))
    usage (0);


  ppl_getopt_init (&opt, argc, argv);

  while ((rv = ppl_getopt (opt, PSP_SERVER_BASEARGS, &c, &optarg)) == PPL_SUCCESS)
    {
      switch (c)
        {
          case 'f':
            cf_config_file = optarg;
            break;
          case 'd':
            /* to be replaced by strtol */
            cf_debug_level = atoi (optarg);
            break;
          case 'l':
            cf_log_file = optarg;
            break;
          case 'v':
            printf ("Server version:     %s\n", PSP_SERVER_BASEVERSION);
            printf ("Server built:       %s\n", server_built);
            partysip_exit (0);
          case 'V':
            show_compile_settings ();
            partysip_exit (0);
          case 'i':
            interactive_mode = 1;
            break;
          case 'p':
            printf ("Show compiled in modules -NOT IMPLEMENTED-\n");
            /*      psp_core_show_compiled_in_plugins(); */
            partysip_exit (0);
          case 'h':
          case '?':
            printf ("\nServer version:     %s\n", PSP_SERVER_BASEVERSION);
            printf ("Server version:     %s\n", PSP_SERVER_BASEVERSION);
            printf ("Server built:       %s\n", server_built);
            usage (0);
          default:
            /* bad cmdline option?  then we die */
            usage (1);
        }
    }

  if (rv != PPL_EOF)
    {
      usage (1);
    }


  /*********************************/
  /*   Load Config File            */

  i = psp_config_load ((char *) cf_config_file);
  if (i != 0)
    {
      printf ("Config name:        %s\n", cf_config_file);
      perror ("ERROR: Could not open config file");
      usage (1);
    }

  /* check some mandatory element in the config file */
  {
    char *serverip = psp_config_get_element ("serverip");
    char *servername = psp_config_get_element ("servername");
    char *serverrealm = psp_config_get_element ("serverrealm");
    char *magicstring = psp_config_get_element ("magicstring");

#ifdef __linux
    if (servername == NULL || serverip == NULL)
      {
        char *tmp_name, *tmp_ip;

        i = ppl_dns_get_local_fqdn (&tmp_name, &tmp_ip);
        if (i == PPL_SUCCESS)
          {
            if (servername == NULL)
              {
                servername = tmp_name;
                psp_config_add_element (sgetcopy ("servername"),
                                        sgetcopy (servername));
                //printf("Autodetected servername: %s\n",servername);
              }
            if (serverip == NULL)
              {
                serverip = tmp_ip;
                psp_config_add_element (sgetcopy ("serverip"),
                                        sgetcopy (serverip));
                //printf("Autodetected serverip: %s\n",serverip);
              }
          }
      }
#endif

    if (serverip == NULL)
      {
        printf ("Bad configuration: \"serverip\" is mandatory\n");
        partysip_exit (1);
      }
    if (servername == NULL)
      {
        printf ("Bad configuration: \"servername\" is mandatory\n");
        partysip_exit (1);
      }
    if (magicstring == NULL || strlen (magicstring) < 10)
      {
        printf
          ("Bad configuration: \"magicstring\" is mandatory and MUST be at least 10 char\n");
        partysip_exit (1);
      }
    if (serverrealm == NULL)
      {
        printf ("Bad configuration: \"serverrealm\" is mandatory.\n");
        partysip_exit (1);
      }

  }


  if (cf_debug_level > 0 && cf_log_file == NULL)
    {
      cf_log_file = psp_config_get_element ("log");
    }

  if (cf_debug_level > 0)
    {
      char *atmp;

      printf ("Server:             %s\n", PSP_SERVER_WHO_AM_I);
      printf ("Debug level:        %i\n", cf_debug_level);
      printf ("Config name:        %s\n", cf_config_file);
      atmp = psp_config_get_element ("serverip");
      printf ("ServerIP:           %s\n", atmp);
      atmp = psp_config_get_element ("servername");
      printf ("ServerName:         %s\n", atmp);
      atmp = psp_config_get_element ("serverrealm");
      printf ("ServerRealm:        '%s'\n", atmp);
      if (cf_log_file == NULL)
        printf ("Log name:           Standard output\n");
      else
        printf ("Log name:           %s\n", cf_log_file);
    }


  /*********************************/
  /* INIT Log File and Log LEVEL   */

  if (cf_debug_level > 0)
    {
      if (cf_log_file != NULL)
        {
          log_file = fopen (cf_log_file, "w+");
      } else
        log_file = stdout;
      if (NULL == log_file)
        {
          printf ("Log name:           %s\n", cf_log_file);
          perror ("ERROR: Could not open log file");
          exit (1);
        }
      TRACE_INITIALIZE (cf_debug_level, log_file);
    }

  sfree ((void *) (opt->argv));
  sfree ((void *) opt);

  i = psp_core_init ();
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not initialize partysip!\n");
      goto main_error;
    }

  /* load general user configuration */
  i = psp_utils_load_users ();
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load user config!\n");
      partysip_exit (0);
    }

  /* from here, let's load plugins */
  i = main_load_plugins ();
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not load plugins!\n");
      partysip_exit (0);
    }

  /* start all modules that we want as thread */
  i = psp_core_start ();
  if (i != 0)
    {
      fprintf (stderr, "ERROR: Could not start modules of partysip!\n");
      goto main_error;
    }

  if (interactive_mode == 1)
    main_run ();
  else
    select (0, NULL, NULL, NULL, NULL);

  partysip_exit (0);
  return 0;
main_error:
  partysip_exit (0);
  return -1;
}
